#!/bin/bash

# make the script exit whenever an error occurs
# set -e

uploaded_file=$1

if [ "$#" -ne 1 ]; then
  echo 'Please provide an input file as follows:'
  echo 'a) ./spectrus.sh set_of_pdb_structures.zip '
  echo 'b) ./spectrus.sh VMD_xyz_trajectory.xyz '
  echo 'c) ./spectrus.sh single_structure.pdb'
  echo 'If a single structure is provided, an ensemble of '
  echo 'configurations is generated from a beta Gaussian Model.'
  echo 'For further details, see the README file.'
  exit 1
fi

if [ ! -f $uploaded_file ]; then
  echo 'File not found'
  exit 1
fi

# find the extension of the loaded file
filename=$(basename "$uploaded_file")
extension="${filename##*.}"

# create the results directory
dirPath=results_${filename%.*}
mkdir -p "$dirPath"
if [ "$?" != 0 ]; then
  echo 'ERROR: directory' $dirPath 'already exists.'
  echo 'Please move or delete the existing directory.'
  echo `date`
  exit 1
fi

# check for pdb, zip or xyz extension
if [ "$extension" = "pdb" ]; then

  mkdir unzipped_pdb_files
  cp "$uploaded_file" unzipped_pdb_files
  cd unzipped_pdb_files

  # generate trajectory from ENM
  echo 'A single pdb file has been uploaded:'
  echo 'creating a beta Gaussian Model...'

  file="$filename"
  echo "$file" > PDB_FILES.TXT
  cp ../src/PARAMS_BETAGM.DAT .

  # run betaGM
  ../src/betagm_lowmem_slepc/betaGM_lowmem_slepc >> ../${dirPath}/spectrus.log 2>> ../${dirPath}/spectrus.err
  if [ "$?" -ne 0 ]
  then
    echo 'Error while generating the beta Gaussian Model.'
    echo 'Aborted.'
    echo `date`
    cd ..
    exit 1 
  fi

  # generate a metatrajectory from the normal modes
  echo 'generating a set of configurations from normal modes...'
  ../src/ENMmodes2traj/ENMmodes2traj results_${file} ${file} >> ../${dirPath}/spectrus.log 2>> ../${dirPath}/spectrus.err
  if [ "$?" -ne 0 ]
  then
    echo 'Error: unable to generate conformations from ENM normal modes.'
    echo 'Aborted.'
    echo `date`
    cd ..
    rm -r unzipped_pdb_files
    exit 1
  fi

  # output
  mv ENM_trajectory.xyz ../${dirPath}/trajectory.xyz

  # print the number of residues
  num_res=`grep " CA " ${file} | wc -l`
  echo $num_res > ../${dirPath}/num_res.dat
  
  # delete unneeded files
  mv results_${file} ../${dirPath}/ENM_modes
  cd ..

elif [ "$extension" = "zip" ]; then

  echo 'Unzipping ' $uploaded_file '...'
  unzip $uploaded_file -d unzipped_pdb_files
  if [ "$?" != 0 ]; then
    echo 'ERROR: ' $uploaded_file 'is not a valid zip file.'
    echo `date`
    rm -rf unzipped_pdb_files
    exit 1
  fi

  num_file=0
  k_old=0

  cd unzipped_pdb_files

  num_pdb_files=`ls -1 *.pdb | wc -l`

  if [ $num_pdb_files -eq 1 ]; then
    # generate trajectory from ENM
    echo 'A single pdb file has been found:'
    echo 'creating a beta Gaussian Model...'

    # betaGM takes care of retaining only the first model and 
    # discarding alternative atom position from pdb file
    file=`ls *.pdb`
    echo "$file" > PDB_FILES.TXT
    cp ../src/PARAMS_BETAGM.DAT .

    # run betaGM
    ../src/betaGM_lowmem_slepc/betaGM_lowmem_slepc >> ../${dirPath}/spectrus.log 2>> ../${dirPath}/spectrus.err
    if [ "$?" -ne 0 ]
    then
      echo 'Error while generating the beta Gaussian Model.'
      echo 'Aborted.'
      echo `date`
      cd ..
      rm -r unzipped_pdb_files
      exit 1 
    fi

    # generate a metatrajectory from the normal modes
    echo 'generating a set of configurations from normal modes...'
    ../src/ENMmodes2traj/ENMmodes2traj results_${file} ${file} 300 >> ../${dirPath}/spectrus.log 2>> ../${dirPath}/spectrus.err
    if [ "$?" -ne 0 ]
    then
      echo 'Error: unable to generate conformations from ENM normal modes.'
      echo 'Aborted.'
      echo `date`
      cd ..
      rm -r unzipped_pdb_files
      exit 1
    fi

    # output
    mv ENM_trajectory.xyz ../${dirPath}/trajectory.xyz

    # print the number of residues
    num_res=`grep " CA " ${file} | wc -l`
    echo $num_res > ../${dirPath}/num_res.dat
    
    # delete unneded files
    mv results_${file} ../${dirPath}/ENM_modes
    cd ..
    
  else 
    for file in $(ls -1 *.pdb); do

      num_file=`expr $num_file + 1`

      # retain only the first model and the A alternative atom, if present, then print only CA
      awk 'BEGIN{stop_printing=0;} \
      {if($1=="ENDMDL") stop_printing=1; \
      if((stop_printing==0) && ($1=="ATOM") && ((substr($0,17,1)==" ")||(substr($0,17,1)=="A"))) \
      print }' ${file} | grep ' CA ' > ../${file}.CA

      num_res=`cat ../${file}.CA | wc -l `

      # check if files have the same number of residues
      if [ $num_file -ne 1 ]; then
        if [ $num_res -ne $k_old ]; then
          echo 'ERROR: A different number of residues has been found in file ' $file
          echo `date`
          cd ..
          rm -r unzipped_pdb_files
          exit 1
        fi
      fi

      k_old=$num_res

    done
  
  
    cd ..

  
    # print the number of residues
    echo $num_res > ${dirPath}/num_res.dat

    # print the file trajectory.xyz
    cat *.CA > trajectory.CA
    cut -c 31-38 trajectory.CA > x.dat
    cut -c 39-46 trajectory.CA > y.dat
    cut -c 47-54 trajectory.CA > z.dat
    paste x.dat y.dat z.dat > ${dirPath}/trajectory.xyz

    rm -f *.CA x.dat y.dat z.dat
  fi
elif [ "$extension" = "xyz" ]; then
  
  # keep only C atoms
  grep "C" $uploaded_file | awk '{print $2, $3, $4}' > ${dirPath}/trajectory.xyz

  # get the number of residues
  num_res=`head -1 $uploaded_file | awk '{print $1}'`

  # get the number of lines
  num_lines=`cat ${dirPath}/trajectory.xyz | wc -l`

  # check if the number of lines in trajectory file is consistent
  remain=`expr $num_lines % $num_res`
  if [ "$remain" != "0" ]; then
    echo "ERROR: The number of residues found is incorrect."
    echo "Be sure that you retained only CA atoms in the trajectory file!"
    echo `date`
    exit 1
  fi

  # compute the number of frames
  num_frames=`expr $num_lines / $num_res`

  echo "Number of residues found: $num_res " >> ${dirPath}/spectrus.log
  echo "Number of frames found:   $num_frames " >> ${dirPath}/spectrus.log 
  if [ "$num_frames" -le 1 ]
  then
       echo 'At least two frames needed'
       echo 'Aborted.'
       echo `date`
       exit 1
  fi


  echo $num_res >  ${dirPath}/num_res.dat

else
  
  echo 'ERROR: Incorrect file format.'
  echo 'Only files with .pdb, .zip or .xyz extension are accepted!'
  echo `date`
  exit 1

fi

if [ "$num_res" -eq 0 ]; then
    echo 'ERROR: no residues '
    echo `date`
    exit 1
fi


echo "N_ATOMS $num_res" >> ${dirPath}/INPUT_PARAMS.DAT
cat INPUT_PARAMS.DAT >> ${dirPath}/INPUT_PARAMS.DAT

echo "Running Spectrus..."
cd ${dirPath}
../src/spectrus/src/spectrus trajectory.xyz >> spectrus.log 2>> spectrus.err

# echo "Generating output..."
# if [ -d "./results" ]; then
#   cd results
#   bash ../../print_figures.sh >> ${dirPath}/spectrus.log 2>> ${dirPath}/spectrus.err
#   cd ..
# fi

# create a directory containing only the files to be downloaded
mkdir spectrus_results
awk 'BEGIN{printf("# domains    \tquality score\n")} {printf("%d \t\t%f \n", $1, $2)}' results/quality_score.dat > spectrus_results/quality_score.dat
# cp results/*png  spectrus_results/
# cp results/*log results/*err spectrus_results/
if [ "$extension" = "xyz" ]; then
  file=`ls -1 results/final_clusterization_kmed-* | head -1`
else
  file=`find ../unzipped_pdb_files/*pdb -maxdepth 1 -type f | head -1`
fi
../src/merge_pdb_and_domain_index_files/merge_PDB_and_domain_index_files ${file} > /dev/null 
mv subdivision_in_*_domains.pdb subdivision_in_*_domains-labels.dat spectrus_results/
# zip -r --move results.zip spectrus_results

rm -rf results/ num_res.dat INPUT_PARAMS.DAT trajectory.xyz

cd ..

rm -rf unzipped_pdb_files

echo 'Finished'
echo `date`

exit 0
